"""A dummy backend for use in tests.

This backend implements the backend API in the simplest way possible.  It is
used in tests of the frontends.
"""


import pykka

from mopidy import backend
from mopidy.models import Playlist, Ref, SearchResult


def create_proxy(config=None, audio=None):
    return DummyBackend.start(config=config, audio=audio).proxy()


class DummyBackend(pykka.ThreadingActor, backend.Backend):
    def __init__(self, config, audio):
        super().__init__()

        self.library = DummyLibraryProvider(backend=self)
        if audio:
            self.playback = backend.PlaybackProvider(audio=audio, backend=self)
        else:
            self.playback = DummyPlaybackProvider(audio=audio, backend=self)
        self.playlists = DummyPlaylistsProvider(backend=self)

        self.uri_schemes = ["dummy"]


class DummyLibraryProvider(backend.LibraryProvider):
    root_directory = Ref.directory(uri="dummy:/", name="dummy")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dummy_library = []
        self.dummy_get_distinct_result = {}
        self.dummy_get_images_result = {}
        self.dummy_browse_result = {}
        self.dummy_find_exact_result = SearchResult()
        self.dummy_search_result = SearchResult()

    def browse(self, path):
        return self.dummy_browse_result.get(path, [])

    def get_distinct(self, field, query=None):
        return self.dummy_get_distinct_result.get(field, set())

    def get_images(self, uris):
        return self.dummy_get_images_result

    def lookup(self, uri):
        return [t for t in self.dummy_library if uri == t.uri]

    def refresh(self, uri=None):
        pass

    def search(self, query=None, uris=None, exact=False):
        if exact:  # TODO: remove uses of dummy_find_exact_result
            return self.dummy_find_exact_result
        return self.dummy_search_result


class DummyPlaybackProvider(backend.PlaybackProvider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._uri = None
        self._time_position = 0

    def pause(self):
        return True

    def play(self):
        return self._uri and self._uri != "dummy:error"

    def change_track(self, track):
        """Pass a track with URI 'dummy:error' to force failure"""
        self._uri = track.uri
        self._time_position = 0
        return True

    def prepare_change(self):
        pass

    def resume(self):
        return True

    def seek(self, time_position):
        self._time_position = time_position
        return True

    def stop(self):
        self._uri = None
        return True

    def get_time_position(self):
        return self._time_position


class DummyPlaylistsProvider(backend.PlaylistsProvider):
    def __init__(self, backend):
        super().__init__(backend)
        self._playlists = []

    def set_dummy_playlists(self, playlists):
        """For tests using the dummy provider through an actor proxy."""
        self._playlists = playlists

    def as_list(self):
        return [
            Ref.playlist(uri=pl.uri, name=pl.name) for pl in self._playlists
        ]

    def get_items(self, uri):
        playlist = self.lookup(uri)
        if playlist is None:
            return
        return [Ref.track(uri=t.uri, name=t.name) for t in playlist.tracks]

    def lookup(self, uri):
        for playlist in self._playlists:
            if playlist.uri == uri:
                return playlist

    def refresh(self):
        pass

    def create(self, name):
        playlist = Playlist(name=name, uri="dummy:%s" % name)
        self._playlists.append(playlist)
        return playlist

    def delete(self, uri):
        playlist = self.lookup(uri)
        if playlist:
            self._playlists.remove(playlist)

    def save(self, playlist):
        old_playlist = self.lookup(playlist.uri)

        if old_playlist is not None:
            index = self._playlists.index(old_playlist)
            self._playlists[index] = playlist
        else:
            self._playlists.append(playlist)

        return playlist



import pytest

from mopidy_mpris.root import Root


@pytest.fixture
def root(config, core):
    return Root(config, core)


def test_fullscreen_returns_false(root):
    assert root.Fullscreen is False


def test_setting_fullscreen_fails(root):
    root.Fullscreen = True

    assert root.Fullscreen is False


def test_can_set_fullscreen_returns_false(root):
    assert root.CanSetFullscreen is False


def test_can_raise_returns_false(root):
    assert root.CanRaise is False


def test_raise_does_nothing(root):
    root.Raise()


def test_can_quit_returns_false(root):
    assert root.CanQuit is False


def test_quit_does_nothing(root):
    root.Quit()


def test_has_track_list_returns_false(root):
    assert root.HasTrackList is False


def test_identify_is_mopidy(root):
    assert root.Identity == "Mopidy"


def test_desktop_entry_is_blank(root, config):
    assert root.DesktopEntry == ""


def test_supported_uri_schemes_includes_backend_uri_schemes(root):
    assert root.SupportedUriSchemes == ["dummy"]


def test_supported_mime_types_has_hardcoded_entries(root):
    assert root.SupportedMimeTypes == [
        "audio/mpeg",
        "audio/x-ms-wma",
        "audio/x-ms-asf",
        "audio/x-flac",
        "audio/flac",
        "audio/l16;channels=2;rate=44100",
        "audio/l16;rate=44100;channels=2",
    ]



import pytest

from mopidy.core import Core

from tests import dummy_audio, dummy_backend, dummy_mixer


@pytest.fixture
def config():
    return {
        "core": {"max_tracklist_length": 10000},
    }


@pytest.fixture
def audio():
    actor = dummy_audio.create_proxy()
    yield actor
    actor.stop()


@pytest.fixture
def backend(audio):
    actor = dummy_backend.create_proxy(audio=audio)
    yield actor
    actor.stop()


@pytest.fixture
def mixer():
    actor = dummy_mixer.create_proxy()
    yield actor
    actor.stop()


@pytest.fixture
def core(config, backend, mixer, audio):
    actor = Core.start(
        config=config, backends=[backend], mixer=mixer, audio=audio
    ).proxy()
    yield actor
    actor.stop()



"""Implementation of org.mpris.MediaPlayer2.Playlists interface.

https://specifications.freedesktop.org/mpris-spec/2.2/Playlists_Interface.html
"""

import base64
import logging
from typing import Union

from pydbus.generic import signal

from mopidy_mpris.interface import Interface

logger = logging.getLogger(__name__)


class Playlists(Interface):
    """
    <node>
      <interface name="org.mpris.MediaPlayer2.Playlists">
        <method name="ActivatePlaylist">
          <arg name="PlaylistId" type="o" direction="in"/>
        </method>
        <method name="GetPlaylists">
          <arg name="Index" type="u" direction="in"/>
          <arg name="MaxCount" type="u" direction="in"/>
          <arg name="Order" type="s" direction="in"/>
          <arg name="ReverseOrder" type="b" direction="in"/>
          <arg name="Playlists" type="a(oss)" direction="out"/>
        </method>
        <signal name="PlaylistChanged">
          <arg name="Playlist" type="oss"/>
        </signal>
        <property name="PlaylistCount" type="u" access="read"/>
        <property name="Orderings" type="as" access="read"/>
        <property name="ActivePlaylist" type="(b(oss))" access="read"/>
      </interface>
    </node>
    """

    INTERFACE = "org.mpris.MediaPlayer2.Playlists"

    def ActivatePlaylist(self, playlist_id):
        logger.debug(
            "%s.ActivatePlaylist(%r) called", self.INTERFACE, playlist_id
        )
        playlist_uri = get_playlist_uri(playlist_id)
        playlist = self.core.playlists.lookup(playlist_uri).get()
        if playlist and playlist.tracks:
            tl_tracks = self.core.tracklist.add(playlist.tracks).get()
            self.core.playback.play(tlid=tl_tracks[0].tlid).get()

    def GetPlaylists(self, index, max_count, order, reverse):
        logger.debug(
            "%s.GetPlaylists(%r, %r, %r, %r) called",
            self.INTERFACE,
            index,
            max_count,
            order,
            reverse,
        )
        playlists = self.core.playlists.as_list().get()
        if order == "Alphabetical":
            playlists.sort(key=lambda p: p.name, reverse=reverse)
        elif order == "User" and reverse:
            playlists.reverse()
        slice_end = index + max_count
        playlists = playlists[index:slice_end]
        results = [(get_playlist_id(p.uri), p.name, "") for p in playlists]
        return results

    PlaylistChanged = signal()

    @property
    def PlaylistCount(self):
        self.log_trace("Getting %s.PlaylistCount", self.INTERFACE)
        return len(self.core.playlists.as_list().get())

    @property
    def Orderings(self):
        self.log_trace("Getting %s.Orderings", self.INTERFACE)
        return [
            "Alphabetical",  # Order by playlist.name
            "User",  # Don't change order
        ]

    @property
    def ActivePlaylist(self):
        self.log_trace("Getting %s.ActivePlaylist", self.INTERFACE)
        playlist_is_valid = False
        playlist = ("/", "None", "")
        return (playlist_is_valid, playlist)


def get_playlist_id(playlist_uri: Union[str, bytes]) -> str:
    # Only A-Za-z0-9_ is allowed, which is 63 chars, so we can't use
    # base64. Luckily, D-Bus does not limit the length of object paths.
    # Since base32 pads trailing bytes with "=" chars, we need to replace
    # them with an allowed character such as "_".
    if isinstance(playlist_uri, str):
        playlist_uri = playlist_uri.encode()
    encoded_uri = base64.b32encode(playlist_uri).decode().replace("=", "_")
    return "/com/mopidy/playlist/%s" % encoded_uri


def get_playlist_uri(playlist_id: Union[str, bytes]) -> str:
    if isinstance(playlist_id, bytes):
        playlist_id = playlist_id.decode()
    encoded_uri = playlist_id.split("/")[-1].replace("_", "=").encode()
    return base64.b32decode(encoded_uri).decode()



import logging

import pykka

from mopidy.core import CoreListener
from mopidy_mpris.playlists import get_playlist_id
from mopidy_mpris.server import Server

logger = logging.getLogger(__name__)


class MprisFrontend(pykka.ThreadingActor, CoreListener):
    def __init__(self, config, core):
        super().__init__()
        self.config = config
        self.core = core
        self.mpris = None

    def on_start(self):
        try:
            self.mpris = Server(self.config, self.core)
            self.mpris.publish()
        except Exception as e:
            logger.warning("MPRIS frontend setup failed (%s)", e)
            self.stop()

    def on_stop(self):
        logger.debug("Removing MPRIS object from D-Bus connection...")
        if self.mpris:
            self.mpris.unpublish()
            self.mpris = None
        logger.debug("Removed MPRIS object from D-Bus connection")

    def on_event(self, event, **kwargs):
        logger.debug("Received %s event", event)
        if self.mpris is None:
            return
        return super().on_event(event, **kwargs)

    def track_playback_paused(self, tl_track, time_position):
        _emit_properties_changed(self.mpris.player, ["PlaybackStatus"])

    def track_playback_resumed(self, tl_track, time_position):
        _emit_properties_changed(self.mpris.player, ["PlaybackStatus"])

    def track_playback_started(self, tl_track):
        _emit_properties_changed(
            self.mpris.player, ["PlaybackStatus", "Metadata"]
        )

    def track_playback_ended(self, tl_track, time_position):
        _emit_properties_changed(
            self.mpris.player, ["PlaybackStatus", "Metadata"]
        )

    def playback_state_changed(self, old_state, new_state):
        _emit_properties_changed(
            self.mpris.player, ["PlaybackStatus", "Metadata"]
        )

    def tracklist_changed(self):
        pass  # TODO Implement if adding tracklist support

    def playlists_loaded(self):
        _emit_properties_changed(self.mpris.playlists, ["PlaylistCount"])

    def playlist_changed(self, playlist):
        playlist_id = get_playlist_id(playlist.uri)
        self.mpris.playlists.PlaylistChanged(playlist_id, playlist.name, "")

    def playlist_deleted(self, uri):
        _emit_properties_changed(self.mpris.playlists, ["PlaylistCount"])

    def options_changed(self):
        _emit_properties_changed(
            self.mpris.player,
            ["LoopStatus", "Shuffle", "CanGoPrevious", "CanGoNext"],
        )

    def volume_changed(self, volume):
        _emit_properties_changed(self.mpris.player, ["Volume"])

    def mute_changed(self, mute):
        _emit_properties_changed(self.mpris.player, ["Volume"])

    def seeked(self, time_position):
        time_position_in_microseconds = time_position * 1000
        self.mpris.player.Seeked(time_position_in_microseconds)

    def stream_title_changed(self, title):
        _emit_properties_changed(self.mpris.player, ["Metadata"])


def _emit_properties_changed(interface, changed_properties):
    props_with_new_values = [
        (p, getattr(interface, p)) for p in changed_properties
    ]
    interface.PropertiesChanged(
        interface.INTERFACE, dict(props_with_new_values), []
    )



import pytest

from mopidy.audio import PlaybackState
from mopidy.models import Track
from mopidy_mpris.playlists import Playlists


@pytest.fixture
def dummy_playlists(core):
    result = {}

    for name, lm in [("foo", 3000000), ("bar", 2000000), ("baz", 1000000)]:
        pl = core.playlists.create(name).get()
        pl = pl.replace(last_modified=lm)
        result[name] = core.playlists.save(pl).get()

    return result


@pytest.fixture
def playlists(config, core, dummy_playlists):
    return Playlists(config, core)


def test_activate_playlist_appends_tracks_to_tracklist(
    core, playlists, dummy_playlists
):
    core.tracklist.add([Track(uri="dummy:old-a"), Track(uri="dummy:old-b")])
    assert core.tracklist.get_length().get() == 2

    pl = dummy_playlists["baz"]
    pl = pl.replace(
        tracks=[
            Track(uri="dummy:baz-a"),
            Track(uri="dummy:baz-b"),
            Track(uri="dummy:baz-c"),
        ]
    )
    pl = core.playlists.save(pl).get()
    playlist_id = playlists.GetPlaylists(0, 100, "User", False)[2][0]

    playlists.ActivatePlaylist(playlist_id)

    assert core.tracklist.get_length().get() == 5
    assert core.playback.get_state().get() == PlaybackState.PLAYING
    assert core.playback.get_current_track().get() == pl.tracks[0]


def test_activate_empty_playlist_is_harmless(core, playlists):
    assert core.tracklist.get_length().get() == 0
    playlist_id = playlists.GetPlaylists(0, 100, "User", False)[2][0]

    playlists.ActivatePlaylist(playlist_id)

    assert core.tracklist.get_length().get() == 0
    assert core.playback.get_state().get() == PlaybackState.STOPPED
    assert core.playback.get_current_track().get() is None


def test_get_playlists_in_alphabetical_order(playlists):
    result = playlists.GetPlaylists(0, 100, "Alphabetical", False)

    assert result == [
        ("/com/mopidy/playlist/MR2W23LZHJRGC4Q_", "bar", ""),
        ("/com/mopidy/playlist/MR2W23LZHJRGC6Q_", "baz", ""),
        ("/com/mopidy/playlist/MR2W23LZHJTG63Y_", "foo", ""),
    ]


def test_get_playlists_in_reverse_alphabetical_order(playlists):
    result = playlists.GetPlaylists(0, 100, "Alphabetical", True)

    assert len(result) == 3
    assert result[0][1] == "foo"
    assert result[1][1] == "baz"
    assert result[2][1] == "bar"


def test_get_playlists_in_user_order(playlists):
    result = playlists.GetPlaylists(0, 100, "User", False)

    assert len(result) == 3
    assert result[0][1] == "foo"
    assert result[1][1] == "bar"
    assert result[2][1] == "baz"


def test_get_playlists_in_reverse_user_order(playlists):
    result = playlists.GetPlaylists(0, 100, "User", True)

    assert len(result) == 3
    assert result[0][1] == "baz"
    assert result[1][1] == "bar"
    assert result[2][1] == "foo"


def test_get_playlists_slice_on_start_of_list(playlists):
    result = playlists.GetPlaylists(0, 2, "User", False)

    assert len(result) == 2
    assert result[0][1] == "foo"
    assert result[1][1] == "bar"


def test_get_playlists_slice_later_in_list(playlists):
    result = playlists.GetPlaylists(2, 2, "User", False)

    assert len(result) == 1
    assert result[0][1] == "baz"


def test_get_playlist_count_returns_number_of_playlists(playlists):
    assert playlists.PlaylistCount == 3


def test_get_orderings_includes_alpha_modified_and_user(playlists):
    result = playlists.Orderings

    assert "Alphabetical" in result
    assert "Created" not in result
    assert "Modified" not in result
    assert "Played" not in result
    assert "User" in result


def test_get_active_playlist_does_not_return_a_playlist(playlists):
    result = playlists.ActivePlaylist

    valid, playlist = result
    playlist_id, playlist_name, playlist_icon_uri = playlist

    assert valid is False
    assert playlist_id == "/"
    assert playlist_name == "None"
    assert playlist_icon_uri == ""



import logging

import pydbus

from mopidy_mpris.player import Player
from mopidy_mpris.playlists import Playlists
from mopidy_mpris.root import Root

logger = logging.getLogger(__name__)


class Server:
    def __init__(self, config, core):
        self.config = config
        self.core = core

        self.root = Root(config, core)
        self.player = Player(config, core)
        self.playlists = Playlists(config, core)

        self._publication_token = None

    def publish(self):
        bus_type = self.config["mpris"]["bus_type"]
        logger.debug("Connecting to D-Bus %s bus...", bus_type)

        if bus_type == "system":
            bus = pydbus.SystemBus()
        else:
            bus = pydbus.SessionBus()

        logger.info("MPRIS server connected to D-Bus %s bus", bus_type)

        self._publication_token = bus.publish(
            "org.mpris.MediaPlayer2.mopidy",
            ("/org/mpris/MediaPlayer2", self.root),
            ("/org/mpris/MediaPlayer2", self.player),
            ("/org/mpris/MediaPlayer2", self.playlists),
        )

    def unpublish(self):
        if self._publication_token:
            self._publication_token.unpublish()
            self._publication_token = None



from unittest import mock

import pytest

from mopidy.core.playback import PlaybackState
from mopidy.models import Playlist, TlTrack
from mopidy_mpris import frontend as frontend_mod
from mopidy_mpris import player, playlists, root, server


@pytest.fixture
def frontend():
    # As a plain class, not an actor:
    result = frontend_mod.MprisFrontend(config=None, core=None)
    result.mpris = mock.Mock(spec=server.Server)
    result.mpris.root = mock.Mock(spec=root.Root)
    result.mpris.root.INTERFACE = root.Root.INTERFACE
    result.mpris.player = mock.Mock(spec=player.Player)
    result.mpris.player.INTERFACE = player.Player.INTERFACE
    result.mpris.playlists = mock.Mock(spec=playlists.Playlists)
    result.mpris.playlists.INTERFACE = playlists.Playlists.INTERFACE
    return result


def test_track_playback_paused_event_changes_playback_status(frontend):
    frontend.mpris.player.PlaybackStatus = "Paused"

    frontend.track_playback_paused(tl_track=TlTrack(), time_position=0)

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE, {"PlaybackStatus": "Paused"}, []
    )


def test_track_playback_resumed_event_changes_playback_status(frontend):
    frontend.mpris.player.PlaybackStatus = "Playing"

    frontend.track_playback_resumed(tl_track=TlTrack(), time_position=0)

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE, {"PlaybackStatus": "Playing"}, []
    )


def test_track_playback_started_changes_playback_status_and_metadata(frontend):
    frontend.mpris.player.Metadata = "..."
    frontend.mpris.player.PlaybackStatus = "Playing"

    frontend.track_playback_started(tl_track=TlTrack())

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE,
        {"Metadata": "...", "PlaybackStatus": "Playing"},
        [],
    )


def test_track_playback_ended_changes_playback_status_and_metadata(frontend):
    frontend.mpris.player.Metadata = "..."
    frontend.mpris.player.PlaybackStatus = "Stopped"

    frontend.track_playback_ended(tl_track=TlTrack(), time_position=0)

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE,
        {"Metadata": "...", "PlaybackStatus": "Stopped"},
        [],
    )


def test_playback_state_changed_changes_playback_status_and_metadata(frontend):
    frontend.mpris.player.Metadata = "..."
    frontend.mpris.player.PlaybackStatus = "Stopped"

    frontend.playback_state_changed(
        PlaybackState.PLAYING, PlaybackState.STOPPED
    )

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE,
        {"Metadata": "...", "PlaybackStatus": "Stopped"},
        [],
    )


def test_playlists_loaded_event_changes_playlist_count(frontend):
    frontend.mpris.playlists.PlaylistCount = 17

    frontend.playlists_loaded()

    frontend.mpris.playlists.PropertiesChanged.assert_called_with(
        playlists.Playlists.INTERFACE, {"PlaylistCount": 17}, []
    )


def test_playlist_changed_event_causes_mpris_playlist_changed_event(frontend):
    playlist = Playlist(uri="dummy:foo", name="foo")

    frontend.playlist_changed(playlist=playlist)

    frontend.mpris.playlists.PlaylistChanged.assert_called_with(
        "/com/mopidy/playlist/MR2W23LZHJTG63Y_", "foo", ""
    )


def test_playlist_deleted_event_changes_playlist_count(frontend):
    frontend.mpris.playlists.PlaylistCount = 17

    frontend.playlist_deleted("dummy:foo")

    frontend.mpris.playlists.PropertiesChanged.assert_called_with(
        playlists.Playlists.INTERFACE, {"PlaylistCount": 17}, []
    )


def test_options_changed_event_changes_loopstatus_and_shuffle(frontend):
    frontend.mpris.player.CanGoPrevious = False
    frontend.mpris.player.CanGoNext = True
    frontend.mpris.player.LoopStatus = "Track"
    frontend.mpris.player.Shuffle = True

    frontend.options_changed()

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE,
        {
            "LoopStatus": "Track",
            "Shuffle": True,
            "CanGoPrevious": False,
            "CanGoNext": True,
        },
        [],
    )


def test_volume_changed_event_changes_volume(frontend):
    frontend.mpris.player.Volume = 1.0

    frontend.volume_changed(volume=100)

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE, {"Volume": 1.0}, []
    )


def test_mute_changed_event_changes_volume(frontend):
    frontend.mpris.player.Volume = 0.0

    frontend.mute_changed(True)

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE, {"Volume": 0.0}, []
    )


def test_seeked_event_causes_mpris_seeked_event(frontend):
    frontend.seeked(time_position=31000)

    frontend.mpris.player.Seeked.assert_called_with(31000000)


def test_stream_title_changed_changes_metadata(frontend):
    frontend.mpris.player.Metadata = "..."

    frontend.stream_title_changed("a new title")

    frontend.mpris.player.PropertiesChanged.assert_called_with(
        player.Player.INTERFACE, {"Metadata": "..."}, []
    )



"""Implementation of org.mpris.MediaPlayer2.Player interface.

https://specifications.freedesktop.org/mpris-spec/2.2/Player_Interface.html
"""


import logging

from gi.repository.GLib import Variant
from pydbus.generic import signal

from mopidy.core import PlaybackState
from mopidy_mpris.interface import Interface

logger = logging.getLogger(__name__)


class Player(Interface):
    """
    <node>
      <interface name="org.mpris.MediaPlayer2.Player">
        <method name="Next"/>
        <method name="Previous"/>
        <method name="Pause"/>
        <method name="PlayPause"/>
        <method name="Stop"/>
        <method name="Play"/>
        <method name="Seek">
          <arg name="Offset" type="x" direction="in"/>
        </method>
        <method name="SetPosition">
          <arg name="TrackId" type="o" direction="in"/>
          <arg name="Position" type="x" direction="in"/>
        </method>
        <method name="OpenUri">
          <arg name="Uri" type="s" direction="in"/>
        </method>
        <signal name="Seeked">
          <arg name="Position" type="x"/>
        </signal>
        <property name="PlaybackStatus" type="s" access="read"/>
        <property name="LoopStatus" type="s" access="readwrite"/>
        <property name="Rate" type="d" access="readwrite"/>
        <property name="Shuffle" type="b" access="readwrite"/>
        <property name="Metadata" type="a{sv}" access="read"/>
        <property name="Volume" type="d" access="readwrite"/>
        <property name="Position" type="x" access="read"/>
        <property name="MinimumRate" type="d" access="read"/>
        <property name="MaximumRate" type="d" access="read"/>
        <property name="CanGoNext" type="b" access="read"/>
        <property name="CanGoPrevious" type="b" access="read"/>
        <property name="CanPlay" type="b" access="read"/>
        <property name="CanPause" type="b" access="read"/>
        <property name="CanSeek" type="b" access="read"/>
        <property name="CanControl" type="b" access="read"/>
      </interface>
    </node>
    """

    INTERFACE = "org.mpris.MediaPlayer2.Player"

    # To override from tests.
    _CanControl = True

    def Next(self):
        logger.debug("%s.Next called", self.INTERFACE)
        if not self.CanGoNext:
            logger.debug("%s.Next not allowed", self.INTERFACE)
            return
        self.core.playback.next().get()

    def Previous(self):
        logger.debug("%s.Previous called", self.INTERFACE)
        if not self.CanGoPrevious:
            logger.debug("%s.Previous not allowed", self.INTERFACE)
            return
        self.core.playback.previous().get()

    def Pause(self):
        logger.debug("%s.Pause called", self.INTERFACE)
        if not self.CanPause:
            logger.debug("%s.Pause not allowed", self.INTERFACE)
            return
        self.core.playback.pause().get()

    def PlayPause(self):
        logger.debug("%s.PlayPause called", self.INTERFACE)
        if not self.CanPause:
            logger.debug("%s.PlayPause not allowed", self.INTERFACE)
            return
        state = self.core.playback.get_state().get()
        if state == PlaybackState.PLAYING:
            self.core.playback.pause().get()
        elif state == PlaybackState.PAUSED:
            self.core.playback.resume().get()
        elif state == PlaybackState.STOPPED:
            self.core.playback.play().get()

    def Stop(self):
        logger.debug("%s.Stop called", self.INTERFACE)
        if not self.CanControl:
            logger.debug("%s.Stop not allowed", self.INTERFACE)
            return
        self.core.playback.stop().get()

    def Play(self):
        logger.debug("%s.Play called", self.INTERFACE)
        if not self.CanPlay:
            logger.debug("%s.Play not allowed", self.INTERFACE)
            return
        state = self.core.playback.get_state().get()
        if state == PlaybackState.PAUSED:
            self.core.playback.resume().get()
        else:
            self.core.playback.play().get()

    def Seek(self, offset):
        logger.debug("%s.Seek called", self.INTERFACE)
        if not self.CanSeek:
            logger.debug("%s.Seek not allowed", self.INTERFACE)
            return
        offset_in_milliseconds = offset // 1000
        current_position = self.core.playback.get_time_position().get()
        new_position = current_position + offset_in_milliseconds
        if new_position < 0:
            new_position = 0
        self.core.playback.seek(new_position).get()

    def SetPosition(self, track_id, position):
        logger.debug("%s.SetPosition called", self.INTERFACE)
        if not self.CanSeek:
            logger.debug("%s.SetPosition not allowed", self.INTERFACE)
            return
        position = position // 1000
        current_tl_track = self.core.playback.get_current_tl_track().get()
        if current_tl_track is None:
            return
        if track_id != get_track_id(current_tl_track.tlid):
            return
        if position < 0:
            return
        if current_tl_track.track.length < position:
            return
        self.core.playback.seek(position).get()

    def OpenUri(self, uri):
        logger.debug("%s.OpenUri called", self.INTERFACE)
        if not self.CanControl:
            # NOTE The spec does not explicitly require this check, but
            # guarding the other methods doesn't help much if OpenUri is open
            # for use.
            logger.debug("%s.OpenUri not allowed", self.INTERFACE)
            return
        # NOTE Check if URI has MIME type known to the backend, if MIME support
        # is added to the backend.
        tl_tracks = self.core.tracklist.add(uris=[uri]).get()
        if tl_tracks:
            self.core.playback.play(tlid=tl_tracks[0].tlid).get()
        else:
            logger.debug('Track with URI "%s" not found in library.', uri)

    Seeked = signal()

    @property
    def PlaybackStatus(self):
        self.log_trace("Getting %s.PlaybackStatus", self.INTERFACE)
        state = self.core.playback.get_state().get()
        if state == PlaybackState.PLAYING:
            return "Playing"
        elif state == PlaybackState.PAUSED:
            return "Paused"
        elif state == PlaybackState.STOPPED:
            return "Stopped"

    @property
    def LoopStatus(self):
        self.log_trace("Getting %s.LoopStatus", self.INTERFACE)
        repeat = self.core.tracklist.get_repeat().get()
        single = self.core.tracklist.get_single().get()
        if not repeat:
            return "None"
        else:
            if single:
                return "Track"
            else:
                return "Playlist"

    @LoopStatus.setter
    def LoopStatus(self, value):
        if not self.CanControl:
            logger.debug("Setting %s.LoopStatus not allowed", self.INTERFACE)
            return
        logger.debug("Setting %s.LoopStatus to %s", self.INTERFACE, value)
        if value == "None":
            self.core.tracklist.set_repeat(False)
            self.core.tracklist.set_single(False)
        elif value == "Track":
            self.core.tracklist.set_repeat(True)
            self.core.tracklist.set_single(True)
        elif value == "Playlist":
            self.core.tracklist.set_repeat(True)
            self.core.tracklist.set_single(False)

    @property
    def Rate(self):
        self.log_trace("Getting %s.Rate", self.INTERFACE)
        return 1.0

    @Rate.setter
    def Rate(self, value):
        if not self.CanControl:
            # NOTE The spec does not explicitly require this check, but it was
            # added to be consistent with all the other property setters.
            logger.debug("Setting %s.Rate not allowed", self.INTERFACE)
            return
        logger.debug("Setting %s.Rate to %s", self.INTERFACE, value)
        if value == 0:
            self.Pause()

    @property
    def Shuffle(self):
        self.log_trace("Getting %s.Shuffle", self.INTERFACE)
        return self.core.tracklist.get_random().get()

    @Shuffle.setter
    def Shuffle(self, value):
        if not self.CanControl:
            logger.debug("Setting %s.Shuffle not allowed", self.INTERFACE)
            return
        logger.debug("Setting %s.Shuffle to %s", self.INTERFACE, value)
        self.core.tracklist.set_random(bool(value))

    @property
    def Metadata(self):
        self.log_trace("Getting %s.Metadata", self.INTERFACE)
        current_tl_track = self.core.playback.get_current_tl_track().get()
        stream_title = self.core.playback.get_stream_title().get()
        if current_tl_track is None:
            return {}
        else:
            (tlid, track) = current_tl_track
            track_id = get_track_id(tlid)
            res = {"mpris:trackid": Variant("o", track_id)}
            if track.length:
                res["mpris:length"] = Variant("x", track.length * 1000)
            if track.uri:
                res["xesam:url"] = Variant("s", track.uri)
            if stream_title or track.name:
                res["xesam:title"] = Variant("s", stream_title or track.name)
            if track.artists:
                artists = list(track.artists)
                artists.sort(key=lambda a: a.name or "")
                res["xesam:artist"] = Variant(
                    "as", [a.name for a in artists if a.name]
                )
            if track.album and track.album.name:
                res["xesam:album"] = Variant("s", track.album.name)
            if track.album and track.album.artists:
                artists = list(track.album.artists)
                artists.sort(key=lambda a: a.name or "")
                res["xesam:albumArtist"] = Variant(
                    "as", [a.name for a in artists if a.name]
                )
            art_url = self._get_art_url(track)
            if art_url:
                res["mpris:artUrl"] = Variant("s", art_url)
            if track.disc_no:
                res["xesam:discNumber"] = Variant("i", track.disc_no)
            if track.track_no:
                res["xesam:trackNumber"] = Variant("i", track.track_no)
            return res

    def _get_art_url(self, track):
        images = self.core.library.get_images([track.uri]).get()
        if images[track.uri]:
            largest_image = sorted(
                images[track.uri], key=lambda i: i.width, reverse=True
            )[0]
            return largest_image.uri

    @property
    def Volume(self):
        self.log_trace("Getting %s.Volume", self.INTERFACE)
        mute = self.core.mixer.get_mute().get()
        volume = self.core.mixer.get_volume().get()
        if volume is None or mute is True:
            return 0
        return volume / 100.0

    @Volume.setter
    def Volume(self, value):
        if not self.CanControl:
            logger.debug("Setting %s.Volume not allowed", self.INTERFACE)
            return
        logger.debug("Setting %s.Volume to %s", self.INTERFACE, value)
        if value is None:
            return
        if value < 0:
            value = 0
        elif value > 1:
            value = 1
        self.core.mixer.set_volume(int(value * 100))
        if value > 0:
            self.core.mixer.set_mute(False)

    @property
    def Position(self):
        self.log_trace("Getting %s.Position", self.INTERFACE)
        return self.core.playback.get_time_position().get() * 1000

    MinimumRate = 1.0
    MaximumRate = 1.0

    @property
    def CanGoNext(self):
        self.log_trace("Getting %s.CanGoNext", self.INTERFACE)
        if not self.CanControl:
            return False
        current_tlid = self.core.playback.get_current_tlid().get()
        next_tlid = self.core.tracklist.get_next_tlid().get()
        return next_tlid != current_tlid

    @property
    def CanGoPrevious(self):
        self.log_trace("Getting %s.CanGoPrevious", self.INTERFACE)
        if not self.CanControl:
            return False
        current_tlid = self.core.playback.get_current_tlid().get()
        previous_tlid = self.core.tracklist.get_previous_tlid().get()
        return previous_tlid != current_tlid

    @property
    def CanPlay(self):
        self.log_trace("Getting %s.CanPlay", self.INTERFACE)
        if not self.CanControl:
            return False
        current_tlid = self.core.playback.get_current_tlid().get()
        next_tlid = self.core.tracklist.get_next_tlid().get()
        return current_tlid is not None or next_tlid is not None

    @property
    def CanPause(self):
        self.log_trace("Getting %s.CanPause", self.INTERFACE)
        if not self.CanControl:
            return False
        # NOTE Should be changed to vary based on capabilities of the current
        # track if Mopidy starts supporting non-seekable media, like streams.
        return True

    @property
    def CanSeek(self):
        self.log_trace("Getting %s.CanSeek", self.INTERFACE)
        if not self.CanControl:
            return False
        # NOTE Should be changed to vary based on capabilities of the current
        # track if Mopidy starts supporting non-seekable media, like streams.
        return True

    @property
    def CanControl(self):
        # NOTE This could be a setting for the end user to change.
        return self._CanControl


def get_track_id(tlid):
    return "/com/mopidy/track/%d" % tlid


def get_track_tlid(track_id):
    assert track_id.startswith("/com/mopidy/track/")
    return track_id.split("/")[-1]



import pytest
from gi.repository import GLib

from mopidy.core import PlaybackState
from mopidy.models import Album, Artist, Image, Track
from mopidy_mpris.player import Player

PLAYING = PlaybackState.PLAYING
PAUSED = PlaybackState.PAUSED
STOPPED = PlaybackState.STOPPED


@pytest.fixture
def player(config, core):
    return Player(config, core)


@pytest.mark.parametrize(
    "state, expected",
    [(PLAYING, "Playing"), (PAUSED, "Paused"), (STOPPED, "Stopped")],
)
def test_get_playback_status(core, player, state, expected):
    core.playback.set_state(state)

    assert player.PlaybackStatus == expected


@pytest.mark.parametrize(
    "repeat, single, expected",
    [
        (False, False, "None"),
        (False, True, "None"),
        (True, False, "Playlist"),
        (True, True, "Track"),
    ],
)
def test_get_loop_status(core, player, repeat, single, expected):
    core.tracklist.set_repeat(repeat)
    core.tracklist.set_single(single)

    assert player.LoopStatus == expected


@pytest.mark.parametrize(
    "status, expected_repeat, expected_single",
    [("None", False, False), ("Track", True, True), ("Playlist", True, False)],
)
def test_set_loop_status(
    core, player, status, expected_repeat, expected_single
):
    player.LoopStatus = status

    assert core.tracklist.get_repeat().get() is expected_repeat
    assert core.tracklist.get_single().get() is expected_single


def test_set_loop_status_is_ignored_if_can_control_is_false(core, player):
    player._CanControl = False
    core.tracklist.set_repeat(True)
    core.tracklist.set_single(True)

    player.LoopStatus = "None"

    assert core.tracklist.get_repeat().get() is True
    assert core.tracklist.get_single().get() is True


def test_get_rate_is_greater_or_equal_than_minimum_rate(player):
    assert player.Rate >= player.MinimumRate


def test_get_rate_is_less_or_equal_than_maximum_rate(player):
    assert player.Rate <= player.MaximumRate


def test_set_rate_to_zero_pauses_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.Rate = 0

    assert core.playback.get_state().get() == PAUSED


def test_set_rate_is_ignored_if_can_control_is_false(core, player):
    player._CanControl = False
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.Rate == 0

    assert core.playback.get_state().get() == PLAYING


@pytest.mark.parametrize("random", [True, False])
def test_get_shuffle(core, player, random):
    core.tracklist.set_random(random)

    assert player.Shuffle is random


@pytest.mark.parametrize("value", [True, False])
def test_set_shuffle(core, player, value):
    core.tracklist.set_random(not value)
    assert core.tracklist.get_random().get() is not value

    player.Shuffle = value

    assert core.tracklist.get_random().get() is value


def test_set_shuffle_is_ignored_if_can_control_is_false(core, player):
    player._CanControl = False
    core.tracklist.set_random(False)

    player.Shuffle = True

    assert core.tracklist.get_random().get() is False


def test_get_metadata_is_empty_when_no_current_track(player):
    assert player.Metadata == {}


def test_get_metadata(core, player):
    core.tracklist.add(
        [
            Track(
                uri="dummy:a",
                length=3600000,
                name="a",
                artists=[Artist(name="b"), Artist(name="c"), Artist(name=None)],
                album=Album(
                    name="d", artists=[Artist(name="e"), Artist(name=None)],
                ),
            )
        ]
    )
    core.playback.play().get()

    (tlid, track) = core.playback.get_current_tl_track().get()

    result = player.Metadata

    assert result["mpris:trackid"] == GLib.Variant(
        "o", "/com/mopidy/track/%d" % tlid
    )
    assert result["mpris:length"] == GLib.Variant("x", 3600000000)
    assert result["xesam:url"] == GLib.Variant("s", "dummy:a")
    assert result["xesam:title"] == GLib.Variant("s", "a")
    assert result["xesam:artist"] == GLib.Variant("as", ["b", "c"])
    assert result["xesam:album"] == GLib.Variant("s", "d")
    assert result["xesam:albumArtist"] == GLib.Variant("as", ["e"])


def test_get_metadata_prefers_stream_title_over_track_name(audio, core, player):
    core.tracklist.add([Track(uri="dummy:a", name="Track name",)])
    core.playback.play().get()

    result = player.Metadata
    assert result["xesam:title"] == GLib.Variant("s", "Track name")

    audio.trigger_fake_tags_changed(
        {
            "organization": [
                "Required for Mopidy core to care about the title"
            ],
            "title": ["Stream title"],
        }
    ).get()

    result = player.Metadata
    assert result["xesam:title"] == GLib.Variant("s", "Stream title")


def test_get_metadata_use_library_image_as_art_url(backend, core, player):
    backend.library.dummy_get_images_result = {
        "dummy:a": [
            Image(uri="http://example.com/small.jpg", width=100, height=100),
            Image(uri="http://example.com/large.jpg", width=200, height=200),
        ],
    }
    core.tracklist.add([Track(uri="dummy:a")])
    core.playback.play().get()

    result = player.Metadata

    assert result["mpris:artUrl"] == GLib.Variant(
        "s", "http://example.com/large.jpg"
    )


def test_get_metadata_has_disc_number_in_album(core, player):
    core.tracklist.add([Track(uri="dummy:a", disc_no=2)])
    core.playback.play().get()

    assert player.Metadata["xesam:discNumber"] == GLib.Variant("i", 2)


def test_get_metadata_has_track_number_in_album(core, player):
    core.tracklist.add([Track(uri="dummy:a", track_no=7)])
    core.playback.play().get()

    assert player.Metadata["xesam:trackNumber"] == GLib.Variant("i", 7)


def test_get_volume_should_return_volume_between_zero_and_one(core, player):
    # dummy_mixer starts out with None as the volume
    assert player.Volume == 0

    core.mixer.set_volume(0)
    assert player.Volume == 0

    core.mixer.set_volume(50)
    assert player.Volume == 0.5

    core.mixer.set_volume(100)
    assert player.Volume == 1


def test_get_volume_should_return_0_if_muted(core, player):
    assert player.Volume == 0

    core.mixer.set_volume(100)
    assert player.Volume == 1

    core.mixer.set_mute(True)
    assert player.Volume == 0

    core.mixer.set_mute(False)
    assert player.Volume == 1


@pytest.mark.parametrize(
    "volume, expected", [(-1.0, 0), (0, 0), (0.5, 50), (1.0, 100), (2.0, 100)]
)
def test_set_volume(core, player, volume, expected):
    player.Volume = volume

    assert core.mixer.get_volume().get() == expected


def test_set_volume_to_not_a_number_does_not_change_volume(core, player):
    core.mixer.set_volume(10).get()

    player.Volume = None

    assert core.mixer.get_volume().get() == 10


def test_set_volume_is_ignored_if_can_control_is_false(core, player):
    player._CanControl = False
    core.mixer.set_volume(0)

    player.Volume = 1.0

    assert core.mixer.get_volume().get() == 0


def test_set_volume_to_positive_value_unmutes_if_muted(core, player):
    core.mixer.set_volume(10).get()
    core.mixer.set_mute(True).get()

    player.Volume = 1.0

    assert core.mixer.get_volume().get() == 100
    assert core.mixer.get_mute().get() is False


def test_set_volume_to_zero_does_not_unmute_if_muted(core, player):
    core.mixer.set_volume(10).get()
    core.mixer.set_mute(True).get()

    player.Volume = 0.0

    assert core.mixer.get_volume().get() == 0
    assert core.mixer.get_mute().get() is True


def test_get_position_returns_time_position_in_microseconds(core, player):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()
    core.playback.seek(10000).get()

    result_in_microseconds = player.Position

    result_in_milliseconds = result_in_microseconds // 1000
    assert result_in_milliseconds >= 10000


def test_get_position_when_no_current_track_should_be_zero(player):
    result_in_microseconds = player.Position

    result_in_milliseconds = result_in_microseconds // 1000
    assert result_in_milliseconds == 0


def test_get_minimum_rate_is_one_or_less(player):
    assert player.MinimumRate <= 1.0


def test_get_maximum_rate_is_one_or_more(player):
    assert player.MaximumRate >= 1.0


def test_can_go_next_is_true_if_can_control_and_other_next_track(core, player):
    player._CanControl = True
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()

    assert player.CanGoNext


def test_can_go_next_is_false_if_next_track_is_the_same(core, player):
    player._CanControl = True
    core.tracklist.add([Track(uri="dummy:a")])
    core.tracklist.set_repeat(True)
    core.playback.play().get()

    assert not player.CanGoNext


def test_can_go_next_is_false_if_can_control_is_false(core, player):
    player._CanControl = False
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()

    assert not player.CanGoNext


def test_can_go_previous_is_true_if_can_control_and_previous_track(
    core, player
):
    player._CanControl = True
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.next().get()

    assert player.CanGoPrevious


def test_can_go_previous_is_false_if_previous_track_is_the_same(core, player):
    player._CanControl = True
    core.tracklist.add([Track(uri="dummy:a")])
    core.tracklist.set_repeat(True)
    core.playback.play().get()

    assert not player.CanGoPrevious


def test_can_go_previous_is_false_if_can_control_is_false(core, player):
    player._CanControl = False
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.next().get()

    assert not player.CanGoPrevious


def test_can_play_is_true_if_can_control_and_current_track(core, player):
    player._CanControl = True
    core.tracklist.add([Track(uri="dummy:a")])
    core.playback.play().get()
    assert core.playback.get_current_track().get()

    assert player.CanPlay


def test_can_play_is_false_if_no_current_track(core, player):
    player._CanControl = True
    assert not core.playback.get_current_track().get()

    assert not player.CanPlay


def test_can_play_if_false_if_can_control_is_false(core, player):
    player._CanControl = False

    assert not player.CanPlay


def test_can_pause_is_true_if_can_control_and_track_can_be_paused(core, player):
    player._CanControl = True

    assert player.CanPause


def test_can_pause_if_false_if_can_control_is_false(core, player):
    player._CanControl = False

    assert not player.CanPause


def test_can_seek_is_true_if_can_control_is_true(core, player):
    player._CanControl = True

    assert player.CanSeek


def test_can_seek_is_false_if_can_control_is_false(core, player):
    player._CanControl = False
    result = player.CanSeek
    assert not result


def test_can_control_is_true(core, player):
    result = player.CanControl
    assert result


def test_next_is_ignored_if_can_go_next_is_false(core, player):
    player._CanControl = False
    assert not player.CanGoNext
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_current_track().get().uri == "dummy:a"

    player.Next()

    assert core.playback.get_current_track().get().uri == "dummy:a"


def test_next_when_playing_skips_to_next_track_and_keep_playing(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_current_track().get().uri == "dummy:a"
    assert core.playback.get_state().get() == PLAYING

    player.Next()

    assert core.playback.get_current_track().get().uri == "dummy:b"
    assert core.playback.get_state().get() == PLAYING


def test_next_when_at_end_of_list_should_stop_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.next().get()
    assert core.playback.get_current_track().get().uri == "dummy:b"
    assert core.playback.get_state().get() == PLAYING
    player.Next()
    assert core.playback.get_state().get() == STOPPED


def test_next_when_paused_should_skip_to_next_track_and_stay_paused(
    core, player
):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.pause().get()
    assert core.playback.get_current_track().get().uri == "dummy:a"
    assert core.playback.get_state().get() == PAUSED
    player.Next()
    assert core.playback.get_current_track().get().uri == "dummy:b"
    assert core.playback.get_state().get() == PAUSED


def test_next_when_stopped_skips_to_next_track_and_stay_stopped(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.stop()
    assert core.playback.get_current_track().get().uri == "dummy:a"
    assert core.playback.get_state().get() == STOPPED
    player.Next()
    assert core.playback.get_current_track().get().uri == "dummy:b"
    assert core.playback.get_state().get() == STOPPED


def test_previous_is_ignored_if_can_go_previous_is_false(core, player):
    player._CanControl = False
    assert not player.CanGoPrevious
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.next().get()
    assert core.playback.get_current_track().get().uri == "dummy:b"

    player.Previous()

    assert core.playback.get_current_track().get().uri == "dummy:b"


def test_previous_when_playing_skips_to_prev_track_and_keep_playing(
    core, player
):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.next().get()
    assert core.playback.get_current_track().get().uri == "dummy:b"
    assert core.playback.get_state().get() == PLAYING

    player.Previous()

    assert core.playback.get_current_track().get().uri == "dummy:a"
    assert core.playback.get_state().get() == PLAYING


def test_previous_when_at_start_of_list_should_stop_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_current_track().get().uri == "dummy:a"
    assert core.playback.get_state().get() == PLAYING

    player.Previous()

    assert core.playback.get_state().get() == STOPPED


def test_previous_when_paused_skips_to_previous_track_and_pause(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.next().get()
    core.playback.pause().get()
    assert core.playback.get_current_track().get().uri == "dummy:b"
    assert core.playback.get_state().get() == PAUSED

    player.Previous()

    assert core.playback.get_current_track().get().uri == "dummy:a"
    assert core.playback.get_state().get() == PAUSED


def test_previous_when_stopped_skips_to_previous_track_and_stops(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.next().get()
    core.playback.stop()
    assert core.playback.get_current_track().get().uri == "dummy:b"
    assert core.playback.get_state().get() == STOPPED

    player.Previous()

    assert core.playback.get_current_track().get().uri == "dummy:a"
    assert core.playback.get_state().get() == STOPPED


def test_pause_is_ignored_if_can_pause_is_false(core, player):
    player._CanControl = False
    assert not player.CanPause
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.Pause()

    assert core.playback.get_state().get() == PLAYING


def test_pause_when_playing_should_pause_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.Pause()

    assert core.playback.get_state().get() == PAUSED


def test_pause_when_paused_has_no_effect(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.pause().get()
    assert core.playback.get_state().get() == PAUSED

    player.Pause()

    assert core.playback.get_state().get() == PAUSED


def test_playpause_is_ignored_if_can_pause_is_false(core, player):
    player._CanControl = False
    assert not player.CanPause
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.PlayPause()

    assert core.playback.get_state().get() == PLAYING


def test_playpause_when_playing_should_pause_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.PlayPause()

    assert core.playback.get_state().get() == PAUSED


def test_playpause_when_paused_should_resume_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.pause().get()

    assert core.playback.get_state().get() == PAUSED
    at_pause = core.playback.get_time_position().get()
    assert at_pause >= 0

    player.PlayPause()

    assert core.playback.get_state().get() == PLAYING
    after_pause = core.playback.get_time_position().get()
    assert after_pause >= at_pause


def test_playpause_when_stopped_should_start_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    assert core.playback.get_state().get() == STOPPED

    player.PlayPause()

    assert core.playback.get_state().get() == PLAYING


def test_stop_is_ignored_if_can_control_is_false(core, player):
    player._CanControl = False
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.Stop()

    assert core.playback.get_state().get() == PLAYING


def test_stop_when_playing_should_stop_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING

    player.Stop()

    assert core.playback.get_state().get() == STOPPED


def test_stop_when_paused_should_stop_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.pause().get()
    assert core.playback.get_state().get() == PAUSED

    player.Stop()

    assert core.playback.get_state().get() == STOPPED


def test_play_is_ignored_if_can_play_is_false(core, player):
    player._CanControl = False
    assert not player.CanPlay
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    assert core.playback.get_state().get() == STOPPED

    player.Play()

    assert core.playback.get_state().get() == STOPPED


def test_play_when_stopped_starts_playback(core, player):
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    assert core.playback.get_state().get() == STOPPED

    player.Play()

    assert core.playback.get_state().get() == PLAYING


def test_play_after_pause_resumes_from_same_position(core, player):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()

    before_pause = core.playback.get_time_position().get()
    assert before_pause >= 0

    player.Pause()

    assert core.playback.get_state().get() == PAUSED
    at_pause = core.playback.get_time_position().get()
    assert at_pause >= before_pause

    player.Play()

    assert core.playback.get_state().get() == PLAYING
    after_pause = core.playback.get_time_position().get()
    assert after_pause >= at_pause


def test_play_when_there_is_no_track_has_no_effect(core, player):
    core.tracklist.clear()
    assert core.playback.get_state().get() == STOPPED

    player.Play()

    assert core.playback.get_state().get() == STOPPED


def test_seek_is_ignored_if_can_seek_is_false(core, player):
    player._CanControl = False
    assert not player.CanSeek
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()

    before_seek = core.playback.get_time_position().get()
    assert before_seek >= 0

    milliseconds_to_seek = 10000
    microseconds_to_seek = milliseconds_to_seek * 1000

    player.Seek(microseconds_to_seek)

    after_seek = core.playback.get_time_position().get()
    assert before_seek <= after_seek
    assert after_seek < before_seek + milliseconds_to_seek


def test_seek_seeks_given_microseconds_forward_in_the_current_track(
    core, player
):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()

    before_seek = core.playback.get_time_position().get()
    assert before_seek >= 0

    milliseconds_to_seek = 10000
    microseconds_to_seek = milliseconds_to_seek * 1000

    player.Seek(microseconds_to_seek)

    assert core.playback.get_state().get() == PLAYING

    after_seek = core.playback.get_time_position().get()
    assert after_seek >= before_seek + milliseconds_to_seek


def test_seek_seeks_given_microseconds_backward_if_negative(core, player):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()
    core.playback.seek(20000).get()

    before_seek = core.playback.get_time_position().get()
    assert before_seek >= 20000

    milliseconds_to_seek = -10000
    microseconds_to_seek = milliseconds_to_seek * 1000

    player.Seek(microseconds_to_seek)

    assert core.playback.get_state().get() == PLAYING

    after_seek = core.playback.get_time_position().get()
    assert after_seek >= before_seek + milliseconds_to_seek
    assert after_seek < before_seek


def test_seek_seeks_to_start_of_track_if_new_position_is_negative(core, player):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()
    core.playback.seek(20000).get()

    before_seek = core.playback.get_time_position().get()
    assert before_seek >= 20000

    milliseconds_to_seek = -30000
    microseconds_to_seek = milliseconds_to_seek * 1000

    player.Seek(microseconds_to_seek)

    assert core.playback.get_state().get() == PLAYING

    after_seek = core.playback.get_time_position().get()
    assert after_seek >= before_seek + milliseconds_to_seek
    assert after_seek < before_seek
    assert after_seek >= 0


def test_seek_skips_to_next_track_if_new_position_gt_track_length(core, player):
    core.tracklist.add(
        [Track(uri="dummy:a", length=40000), Track(uri="dummy:b")]
    )
    core.playback.play().get()
    core.playback.seek(20000).get()

    before_seek = core.playback.get_time_position().get()
    assert before_seek >= 20000
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"

    milliseconds_to_seek = 50000
    microseconds_to_seek = milliseconds_to_seek * 1000

    player.Seek(microseconds_to_seek)

    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:b"

    after_seek = core.playback.get_time_position().get()
    assert after_seek >= 0
    assert after_seek < before_seek


def test_set_position_is_ignored_if_can_seek_is_false(core, player):
    player.get_CanSeek = lambda *_: False
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()

    before_set_position = core.playback.get_time_position().get()
    assert before_set_position <= 5000

    track_id = "a"

    position_to_set_in_millisec = 20000
    position_to_set_in_microsec = position_to_set_in_millisec * 1000

    player.SetPosition(track_id, position_to_set_in_microsec)

    after_set_position = core.playback.get_time_position().get()
    assert before_set_position <= after_set_position
    assert after_set_position < position_to_set_in_millisec


def test_set_position_sets_the_current_track_position_in_microsecs(
    core, player
):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()

    before_set_position = core.playback.get_time_position().get()
    assert before_set_position <= 5000
    assert core.playback.get_state().get() == PLAYING

    track_id = "/com/mopidy/track/1"

    position_to_set_in_millisec = 20000
    position_to_set_in_microsec = position_to_set_in_millisec * 1000

    player.SetPosition(track_id, position_to_set_in_microsec)

    assert core.playback.get_state().get() == PLAYING

    after_set_position = core.playback.get_time_position().get()
    assert after_set_position >= position_to_set_in_millisec


def test_set_position_does_nothing_if_the_position_is_negative(core, player):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()
    core.playback.seek(20000)

    before_set_position = core.playback.get_time_position().get()
    assert before_set_position >= 20000
    assert before_set_position <= 25000
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"

    track_id = "/com/mopidy/track/1"

    position_to_set_in_millisec = -1000
    position_to_set_in_microsec = position_to_set_in_millisec * 1000

    player.SetPosition(track_id, position_to_set_in_microsec)

    after_set_position = core.playback.get_time_position().get()
    assert after_set_position >= before_set_position
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"


def test_set_position_does_nothing_if_position_is_gt_track_length(core, player):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()
    core.playback.seek(20000)

    before_set_position = core.playback.get_time_position().get()
    assert before_set_position >= 20000
    assert before_set_position <= 25000
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"

    track_id = "a"

    position_to_set_in_millisec = 50000
    position_to_set_in_microsec = position_to_set_in_millisec * 1000

    player.SetPosition(track_id, position_to_set_in_microsec)

    after_set_position = core.playback.get_time_position().get()
    assert after_set_position >= before_set_position
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"


def test_set_position_is_noop_if_track_id_isnt_current_track(core, player):
    core.tracklist.add([Track(uri="dummy:a", length=40000)])
    core.playback.play().get()
    core.playback.seek(20000)

    before_set_position = core.playback.get_time_position().get()
    assert before_set_position >= 20000
    assert before_set_position <= 25000
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"

    track_id = "b"

    position_to_set_in_millisec = 0
    position_to_set_in_microsec = position_to_set_in_millisec * 1000

    player.SetPosition(track_id, position_to_set_in_microsec)

    after_set_position = core.playback.get_time_position().get()
    assert after_set_position >= before_set_position
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"


def test_open_uri_is_ignored_if_can_control_is_false(backend, core, player):
    player._CanControl = False
    backend.library.dummy_library = [Track(uri="dummy:/test/uri")]

    player.OpenUri("dummy:/test/uri")

    assert core.tracklist.get_length().get() == 0


def test_open_uri_ignores_uris_with_unknown_uri_scheme(backend, core, player):
    assert core.get_uri_schemes().get() == ["dummy"]
    backend.library.dummy_library = [Track(uri="notdummy:/test/uri")]

    player.OpenUri("notdummy:/test/uri")

    assert core.tracklist.get_length().get() == 0


def test_open_uri_adds_uri_to_tracklist(backend, core, player):
    backend.library.dummy_library = [Track(uri="dummy:/test/uri")]

    player.OpenUri("dummy:/test/uri")

    assert core.tracklist.get_length().get() == 1
    assert core.tracklist.get_tracks().get()[0].uri == "dummy:/test/uri"


def test_open_uri_starts_playback_of_new_track_if_stopped(
    backend, core, player
):
    backend.library.dummy_library = [Track(uri="dummy:/test/uri")]
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    assert core.playback.get_state().get() == STOPPED

    player.OpenUri("dummy:/test/uri")

    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:/test/uri"


def test_open_uri_starts_playback_of_new_track_if_paused(backend, core, player):
    backend.library.dummy_library = [Track(uri="dummy:/test/uri")]
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    core.playback.pause().get()
    assert core.playback.get_state().get() == PAUSED
    assert core.playback.get_current_track().get().uri == "dummy:a"

    player.OpenUri("dummy:/test/uri")

    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:/test/uri"


def test_open_uri_starts_playback_of_new_track_if_playing(
    backend, core, player
):
    backend.library.dummy_library = [Track(uri="dummy:/test/uri")]
    core.tracklist.add([Track(uri="dummy:a"), Track(uri="dummy:b")])
    core.playback.play().get()
    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:a"

    player.OpenUri("dummy:/test/uri")

    assert core.playback.get_state().get() == PLAYING
    assert core.playback.get_current_track().get().uri == "dummy:/test/uri"



import pathlib

import pkg_resources

from mopidy import config, exceptions, ext

__version__ = pkg_resources.get_distribution("Mopidy-MPRIS").version


class Extension(ext.Extension):

    dist_name = "Mopidy-MPRIS"
    ext_name = "mpris"
    version = __version__

    def get_default_config(self):
        return config.read(pathlib.Path(__file__).parent / "ext.conf")

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["desktop_file"] = config.Deprecated()
        schema["bus_type"] = config.String(choices=["session", "system"])
        return schema

    def validate_environment(self):
        try:
            import pydbus  # noqa
        except ImportError as e:
            raise exceptions.ExtensionError("pydbus library not found", e)

    def setup(self, registry):
        from .frontend import MprisFrontend

        registry.add("frontend", MprisFrontend)



"""Implementation of org.mpris.MediaPlayer2 interface.

https://specifications.freedesktop.org/mpris-spec/2.2/Media_Player.html
"""


import logging

from mopidy_mpris.interface import Interface

logger = logging.getLogger(__name__)


class Root(Interface):
    """
    <node>
      <interface name="org.mpris.MediaPlayer2">
        <method name="Raise"/>
        <method name="Quit"/>
        <property name="CanQuit" type="b" access="read"/>
        <property name="CanRaise" type="b" access="read"/>
        <property name="Fullscreen" type="b" access="readwrite"/>
        <property name="CanSetFullscreen" type="b" access="read"/>
        <property name="HasTrackList" type="b" access="read"/>
        <property name="Identity" type="s" access="read"/>
        <property name="DesktopEntry" type="s" access="read"/>
        <property name="SupportedUriSchemes" type="as" access="read"/>
        <property name="SupportedMimeTypes" type="as" access="read"/>
      </interface>
    </node>
    """

    INTERFACE = "org.mpris.MediaPlayer2"

    def Raise(self):
        logger.debug("%s.Raise called", self.INTERFACE)
        # Do nothing, as we do not have a GUI

    def Quit(self):
        logger.debug("%s.Quit called", self.INTERFACE)
        # Do nothing, as we do not allow MPRIS clients to shut down Mopidy

    CanQuit = False

    @property
    def Fullscreen(self):
        self.log_trace("Getting %s.Fullscreen", self.INTERFACE)
        return False

    @Fullscreen.setter
    def Fullscreen(self, value):
        logger.debug("Setting %s.Fullscreen to %s", self.INTERFACE, value)
        pass

    CanSetFullscreen = False
    CanRaise = False
    HasTrackList = False  # NOTE Change if adding optional track list support
    Identity = "Mopidy"

    @property
    def DesktopEntry(self):
        self.log_trace("Getting %s.DesktopEntry", self.INTERFACE)
        # This property is optional to expose. If we set this to "mopidy", the
        # basename of "mopidy.desktop", some MPRIS clients will start a new
        # Mopidy instance in a terminal window if one clicks outside the
        # buttons of the UI. This is probably never what the user wants.
        return ""

    @property
    def SupportedUriSchemes(self):
        self.log_trace("Getting %s.SupportedUriSchemes", self.INTERFACE)
        return self.core.get_uri_schemes().get()

    @property
    def SupportedMimeTypes(self):
        # NOTE Return MIME types supported by local backend if support for
        # reporting supported MIME types is added.
        self.log_trace("Getting %s.SupportedMimeTypes", self.INTERFACE)
        return [
            "audio/mpeg",
            "audio/x-ms-wma",
            "audio/x-ms-asf",
            "audio/x-flac",
            "audio/flac",
            "audio/l16;channels=2;rate=44100",
            "audio/l16;rate=44100;channels=2",
        ]



import unittest
from unittest import mock

from mopidy_mpris import Extension
from mopidy_mpris import frontend as frontend_lib


class ExtensionTest(unittest.TestCase):
    def test_get_default_config(self):
        ext = Extension()

        config = ext.get_default_config()

        self.assertIn("[mpris]", config)
        self.assertIn("enabled = true", config)
        self.assertIn("bus_type = session", config)

    def test_get_config_schema(self):
        ext = Extension()

        schema = ext.get_config_schema()

        self.assertIn("desktop_file", schema)
        self.assertIn("bus_type", schema)

    def test_get_frontend_classes(self):
        ext = Extension()
        registry = mock.Mock()

        ext.setup(registry)

        registry.add.assert_called_once_with(
            "frontend", frontend_lib.MprisFrontend
        )



import logging

from pydbus.generic import signal

logger = logging.getLogger(__name__)

# This should be kept in sync with mopidy.internal.log.TRACE_LOG_LEVEL
TRACE_LOG_LEVEL = 5


class Interface:
    def __init__(self, config, core):
        self.config = config
        self.core = core

    PropertiesChanged = signal()

    def log_trace(self, *args, **kwargs):
        logger.log(TRACE_LOG_LEVEL, *args, **kwargs)
